/*******************************************************************
 *
 *	File:		CDeskwinEngine.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 1999, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <Mmsystem.h>
#include "Deskwin.h"
#include "CDeskwinEngine.h"
#include "strings.h"

static UINT GetPreferredAudioOutDevice()
{
	UINT preferredAudioOutputDevice = -1;

	HWAVEOUT		waveOut;
	WAVEFORMATEX	waveFormatEx = {0,};
	
		// simple mono 8KHz PCM
	waveFormatEx.wFormatTag			= WAVE_FORMAT_PCM;
	waveFormatEx.nChannels			= 1;
	waveFormatEx.nSamplesPerSec		= 8000 * 1000; 
	waveFormatEx.wBitsPerSample		= 8; 
	waveFormatEx.nBlockAlign		= waveFormatEx.nChannels * waveFormatEx.wBitsPerSample / 8; 
	waveFormatEx.nAvgBytesPerSec	= waveFormatEx.nSamplesPerSec * waveFormatEx.nBlockAlign; 
	waveFormatEx.cbSize = 0;

	MMRESULT res = ::waveOutOpen(&waveOut, WAVE_MAPPER, &waveFormatEx, 0, 0, 0);
	if (res == MMSYSERR_NOERROR)
	{
		res = mixerGetID((HMIXEROBJ) waveOut, &preferredAudioOutputDevice, MIXER_OBJECTF_HWAVEOUT);
		waveOutClose(waveOut);
	}
	
	return preferredAudioOutputDevice;
}


unsigned long __stdcall HungThreadWatcher(void* aThreadData)
{
	TWindow window = (TWindow) aThreadData;

	DPRINT(("MonitorHungWindows: start, window=0x%x", window));

		/*
		 * Make sure we don't get multiple threads for the same hung window
		 * This easily happens when you keep switching desktops.
		 */
	char	mutexName[32];
	sprintf(mutexName, "DESKWIN_MON_0x%x", window);
	HANDLE h = ::CreateMutex(NULL, true, mutexName);
	if (GetLastError() == ERROR_ALREADY_EXISTS)
	{
		CloseHandle(h);
		DPRINT(("MonitorHungWindows: done: monitor already running for window=0x%x", window));
		return 0;
	}

	while (true)
	{
		if (!::IsWindow(window))
		{
			DPRINT(("MonitorHungWindows: window=0x%x is gone", window));
			break;
		}

		if (!::IsWindowVisible(window))
		{
			DPRINT(("MonitorHungWindows: trying to make window=0x%x visible", window));
			ShowWindowAsync(window, SW_SHOW);
			::Sleep(500);
			
			if (IsWindowVisible(window))
			{
				DPRINT(("MonitorHungWindows: made window=0x%x visible", window));
				break;
			}
			else
			{
				DPRINT(("MonitorHungWindows: unfortunately window=0x%x visible did not become visible", window));
			}
		}
		else
		{
			DPRINT(("MonitorHungWindows: window=0x%x still hung", window));
			::Sleep(3000);
		}
	}
	DPRINT(("MonitorHungWindows: done, window=0x%x", window));

	::CloseHandle(h);
	return 0;
}


static void ChangeVolume(int aDelta)
{
	UINT preferredAudioOutDevice = GetPreferredAudioOutDevice();

	HMIXER	mixer;
	if(::mixerOpen(&mixer, preferredAudioOutDevice, 0, 0, 0 ) != MMSYSERR_NOERROR)
	{
		return;
	}

	MIXERLINE	mixerLine;
	mixerLine.cbStruct = sizeof(MIXERLINE);
	mixerLine.dwComponentType = MIXERLINE_COMPONENTTYPE_DST_SPEAKERS;
	if (::mixerGetLineInfo((HMIXEROBJ) mixer, &mixerLine, MIXER_GETLINEINFOF_COMPONENTTYPE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

	MIXERCONTROL		mixerControl;
	MIXERLINECONTROLS	mixerLineControls;
	mixerLineControls.cbStruct		= sizeof(MIXERLINECONTROLS);
	mixerLineControls.dwLineID		= mixerLine.dwLineID;
	mixerLineControls.cControls		= 1;
	mixerLineControls.dwControlType	= MIXERCONTROL_CONTROLTYPE_VOLUME;
	mixerLineControls.pamxctrl		= &mixerControl;
	mixerLineControls.cbmxctrl		= sizeof(MIXERCONTROL);
	if (::mixerGetLineControls((HMIXEROBJ) mixer, &mixerLineControls, MIXER_GETLINECONTROLSF_ONEBYTYPE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

	MIXERCONTROLDETAILS	mixerControlDetails;
	MIXERCONTROLDETAILS_SIGNED	 mixerControlDetailsSigned;
	mixerControlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	mixerControlDetails.dwControlID = mixerControl.dwControlID;
	mixerControlDetails.cChannels = 1;
	mixerControlDetails.cMultipleItems = 0;
	mixerControlDetails.paDetails = &mixerControlDetailsSigned;
	mixerControlDetails.cbDetails = sizeof(LPMIXERCONTROLDETAILS_SIGNED);
	if (::mixerGetControlDetails((HMIXEROBJ) mixer, &mixerControlDetails, MIXER_GETCONTROLDETAILSF_VALUE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

		// change volume
	aDelta *= (mixerControl.Bounds.lMaximum - mixerControl.Bounds.lMinimum) / 10;
	mixerControlDetailsSigned.lValue += aDelta;
	if (mixerControlDetailsSigned.lValue < 0)
		mixerControlDetailsSigned.lValue = 0;

	if (::mixerSetControlDetails((HMIXEROBJ) mixer, &mixerControlDetails, MIXER_SETCONTROLDETAILSF_VALUE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

	::mixerClose(mixer);
	return;
}





void CDeskwinEngine::VolumeToggleMute()
{
	UINT preferredAudioOutDevice = GetPreferredAudioOutDevice();

	HMIXER	mixer;
	if(::mixerOpen(&mixer, preferredAudioOutDevice, 0, 0, MIXER_OBJECTF_MIXER ) != MMSYSERR_NOERROR)
	{
		return;
	}

	MIXERLINE	mixerLine;
	memset(&mixerLine, 0, sizeof(MIXERLINE));
	mixerLine.cbStruct = sizeof(MIXERLINE);
	mixerLine.dwComponentType = MIXERLINE_COMPONENTTYPE_DST_SPEAKERS;
	MMRESULT res = ::mixerGetLineInfo((HMIXEROBJ) mixer, &mixerLine, MIXER_GETLINEINFOF_COMPONENTTYPE);
	if (res != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

	MIXERCONTROL		mixerControl;
	MIXERLINECONTROLS	mixerLineControls;
	mixerLineControls.cbStruct		= sizeof(MIXERLINECONTROLS);
	mixerLineControls.dwLineID		= mixerLine.dwLineID;
	mixerLineControls.cControls		= 1;
	mixerLineControls.dwControlType	= MIXERCONTROL_CONTROLTYPE_MUTE;
	mixerLineControls.pamxctrl		= &mixerControl;
	mixerLineControls.cbmxctrl		= sizeof(MIXERCONTROL);
	if (::mixerGetLineControls((HMIXEROBJ) mixer, &mixerLineControls, MIXER_GETLINECONTROLSF_ONEBYTYPE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

	MIXERCONTROLDETAILS	mixerControlDetails;
	MIXERCONTROLDETAILS_UNSIGNED	 mixerControlDetailsUnsigned;
	mixerControlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	mixerControlDetails.dwControlID = mixerControl.dwControlID;
	mixerControlDetails.cChannels = 1;
	mixerControlDetails.cMultipleItems = 0;
	mixerControlDetails.paDetails = &mixerControlDetailsUnsigned;
	mixerControlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	if (::mixerGetControlDetails((HMIXEROBJ) mixer, &mixerControlDetails, MIXER_GETCONTROLDETAILSF_VALUE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

		// toggle mute
	mixerControlDetailsUnsigned.dwValue = mixerControlDetailsUnsigned.dwValue ? 0 : 1;

	if (::mixerSetControlDetails((HMIXEROBJ) mixer, &mixerControlDetails, MIXER_SETCONTROLDETAILSF_VALUE) != MMSYSERR_NOERROR)
	{
		::mixerClose(mixer);
		return;
	}

	::mixerClose(mixer);
	return;
}


void CDeskwinEngine::VolumeUp()
{
	ChangeVolume(1);
}

void CDeskwinEngine::VolumeDown()
{
	ChangeVolume(-1);
}



/*
 * Normally, when switching desktops there are only 2 windows that needs
 * special attention: the taskbar & the shell. Although the shell is the one
 * managing the icons on the desktop, there seems to be a special helper
 * window involved when pressing the "clear desktop" button in the taskbar.
 * It seems to temporarily raise a window which looks identical to the desktop
 * window in front of everything else. If you then switch to another *empty* 
 * desktop, that new desktop will not show the desktop icons.
 * Using the excelent SPY tool I noticed that this helper window only seems to
 * responds to WM_ACTIVATEAPP messages.
 * 
 * To cut a long story short, when Capturing the current windows, deskwin
 * will filter out this helper  window and explicitely deactive it by sending
 * it a WM_ACTIVATEAPP saying that an app has been activated.
 */
static bool IsStupidExplorerWorkerWindow(string aClassname)
{
	return aClassname.compare("WorkerW") == 0;
}

static void GetRidOfStupidExplorerWorkerWindow(TWindow aWindow)
{
	::PostMessage(aWindow, WM_ACTIVATEAPP, TRUE, (long) CDeskwinEngine::GetExplorerWindow());
}

/*
 * Windows XP has tool tips popping up all over the show. When one is up and
 * switching desktops back and forth, there is some "shadow" debris on screen.
 * We just filter out those windows and close them. 
 */
static bool IsStupidExplorerFaderWindow(string aClassname)
{
	return (aClassname.compare("SysShadow") == 0);
}

static bool IsToolTipWindow(string aClassname)
{
	return (aClassname.compare(KToolTipClassName) == 0);
}

static void GetRidOfStupidExplorerFaderWindow(TWindow aWindow)
{
	::PostMessage(aWindow, WM_CLOSE, 0, 0);
}


CDeskwinEngine* CDeskwinEngine::iDeskwinEngine = NULL;

CDeskwinEngine* CDeskwinEngine::Instance()
{
	if (!CDeskwinEngine::iDeskwinEngine)
	{
		CDeskwinEngine::iDeskwinEngine = new CDeskwinEngine;
	}
	return CDeskwinEngine::iDeskwinEngine;
}


CDeskwinEngine::CDeskwinEngine() : iCurrentDesktop(0), iNoOfDesktops(1)
{
	RestoreWindowState();
	DeleteStoredWindowState();
	iWindowLists.resize(iNoOfDesktops);
}

CDeskwinEngine::~CDeskwinEngine()
{

}




const TStickyWindowInfo& CDeskwinEngine::StickyWindowInfo()
{
	return iStickyWindowInfo;
}

static TStickyWindow* FindStickyWindowRule(TStickyWindowInfo& aStickyWindowInfo, const string& aTitleName)
{
	for (TStickyWindowInfo::iterator i=aStickyWindowInfo.begin() ; i!=aStickyWindowInfo.end() ; i++)
	{
		TStickyWindow&	stickyWindow = *i;
		if (stickyWindow.iRuleName == aTitleName)
			return &stickyWindow;
	}
	return NULL;
}

static void ToggleDesktopInRule(TStickyWindow& aStickyWindow, int aDesktop, bool aEnable, int aNoOfDesktops)
{
	if (aEnable)
	{
		switch (aStickyWindow.iStickyWindowDesktop)
		{
			case EStickyWindowDesktopAll:
				// nothing
				break;
			case EStickyWindowDesktopInclude:
				aStickyWindow.iDesktopList.remove(aDesktop);
				if (aStickyWindow.iDesktopList.size() == (aNoOfDesktops - 1))
				{
					aStickyWindow.iStickyWindowDesktop = EStickyWindowDesktopAll;
					aStickyWindow.iDesktopList.clear();
				}
				else
				{
					aStickyWindow.iDesktopList.push_back(aDesktop);
				}
				break;
			case EStickyWindowDesktopExclude:
				aStickyWindow.iDesktopList.remove(aDesktop);
				if (aStickyWindow.iDesktopList.size() == 0)
				{
					aStickyWindow.iStickyWindowDesktop = EStickyWindowDesktopAll;
				}
				break;
			default:
				break;
		}
	}
	else
	{
		switch (aStickyWindow.iStickyWindowDesktop)
		{
			case EStickyWindowDesktopAll:
				aStickyWindow.iStickyWindowDesktop = EStickyWindowDesktopExclude;
				aStickyWindow.iDesktopList.push_back(aDesktop);
				break;
			case EStickyWindowDesktopInclude:
				aStickyWindow.iDesktopList.remove(aDesktop);
				break;
			case EStickyWindowDesktopExclude:
				aStickyWindow.iDesktopList.remove(aDesktop);
				aStickyWindow.iDesktopList.push_back(aDesktop);
				break;
			default:
				break;
		}
	}
}

void CDeskwinEngine::HideOrShowWindow(int aDesktop, TWindow aWindow)
{
	bool enable = ::IsWindowVisible(aWindow) ? FALSE : TRUE;

	string		titleName;
	GetWindowName(aWindow, titleName);
	
	TStickyWindow*	stickyWindow = FindStickyWindowRule(iStickyWindowInfo, titleName);
	if (stickyWindow)
	{
		ToggleDesktopInRule(*stickyWindow, aDesktop, enable, iNoOfDesktops);
	}

	MyShowWindow(aWindow, enable);
}


void CDeskwinEngine::MyShowWindow(TWindow aWindow, bool aShow)
{
//	::ShowWindow(aWindow, aShow ? SW_SHOW : SW_HIDE);
	::ShowWindowAsync(aWindow, aShow ? SW_SHOW : SW_HIDE);

	StoreWindowState(aWindow, ::IsWindowVisible(aWindow) ? true:false);
}


void CDeskwinEngine::HideOrShowExplorerOrDesktop(int aDesktop)
{
	HideOrShowWindow(aDesktop, GetExplorerWindow());
	HideOrShowWindow(aDesktop, GetTaskbarWindow());
}


void CDeskwinEngine::HandleStickyWindowInfoChanged()
{
	TWindowList	windowList;
	GetAllWindows(windowList);

	for (int i=0 ; i<iNoOfDesktops ; i++)
	{
		TWindowList	currentWindowList;
		TWindowList	newWindowList;
		
		GetWindowList(i, currentWindowList);

		for (TWindowList::iterator j=windowList.begin() ; j!=windowList.end() ; j++)
		{
			bool	isRuleMatch;
			bool	isSticky;
			int		ignoreRuleNo;
			IsStickyWindow1(i, *j, isRuleMatch, isSticky, ignoreRuleNo);
			if (isRuleMatch)
			{
				if (isSticky)
					newWindowList.push_back(*j);
				if (i == iCurrentDesktop)
				{
					MyShowWindow(*j, isSticky);
				}
			}
			else
			{
				if (find(currentWindowList.begin(), currentWindowList.end(), *j) != currentWindowList.end())
				{
					if ((i==iCurrentDesktop) || !::IsWindowVisible(*j))
					{
						newWindowList.push_back(*j);
					}
				}
			}
		}
		iWindowLists[i] = newWindowList;
	}
}

void CDeskwinEngine::SetStickyWindowInfo(const TStickyWindowInfo& aStickyWindowInfo)
{
	iStickyWindowInfo = aStickyWindowInfo;
	HandleStickyWindowInfoChanged();
}


#define MAX_TEXT	1024

void CDeskwinEngine::GetWindowName(TWindow aWindow, string& aName)
{
	char	text[MAX_TEXT];

	text[0] = 0;

	if (aWindow != GetTaskbarWindow())
	{
		::GetWindowText(aWindow, text, MAX_TEXT);
		text[MAX_TEXT - 1] = 0;
		aName.assign(text); 
	}
	else
	{
		aName.assign("Taskbar"); 
	}
}


void CDeskwinEngine::GetClassName(TWindow aWindow, string& aName)
{
	char	text[MAX_TEXT];

	text[0] = 0;
	::GetClassName(aWindow, text, MAX_TEXT);
	text[MAX_TEXT - 1] = 0;
	aName.assign(text); 
}

TWindow	CDeskwinEngine::GetStartMenuWindow()
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	while (window)
	{
		string	name;
		GetClassName(window, name);
		if (name.compare(KStartMenuClassName) == 0)
			return window;
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}

	return (TWindow) 0;
}

TWindow	CDeskwinEngine::GetTaskbarWindow()
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	while (window)
	{
		string	name;
		GetClassName(window, name);
		if (name.compare(KTaskbarClassName) == 0)
			return window;
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}

	return (TWindow) 0;
}


TWindow	CDeskwinEngine::GetExplorerWindow()
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	while (window)
	{
		string	name;
		GetWindowName(window, name);
		if (name.compare(KExplorerName) == 0)
			return window;
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}

	return (TWindow) 0;
}


void CDeskwinEngine::SetNoOfDesktops(int aNoOfDesktops)
{
	ASSERT(aNoOfDesktops > 0);

	if ((aNoOfDesktops != iNoOfDesktops) && (iNoOfDesktops != 0) )
	{
			/*
			 * If we're going to remove desktops, copy all windows to desktop 0
			 */
		if (aNoOfDesktops < iNoOfDesktops)
		{
				/*
				 * Set new current desktop if it's about to be removed
				 */
			if (iCurrentDesktop >= aNoOfDesktops)
			{
				SwitchTo(aNoOfDesktops - 1);
			}

				/*
				 * Copy all windows to desktop 0
				 */
			int						c;
			TWindowList&			winListDesktop0 = iWindowLists.at(0);
			TWindowLists::iterator	i;

			for (i = iWindowLists.end() - 1, c=iNoOfDesktops; c>aNoOfDesktops ; i--, c--)
			{
				winListDesktop0.insert(winListDesktop0.end(), (*i).begin(), (*i).end());
			}

				/*
				 * Show windows when desktop 0 is visible
				 */
			if (iCurrentDesktop == 0)
			{
				TWindowList  dummy;
				ShowDesktop(winListDesktop0, iActiveWindows[0], dummy);
			}
		}
	}

	int		iOldNoOfDesktops = iNoOfDesktops;
	iNoOfDesktops = aNoOfDesktops;
	iWindowLists.resize(iNoOfDesktops);
	iStickyWindowInfo.resize(iNoOfDesktops);
	iActiveWindows.resize(iNoOfDesktops);

	if (iOldNoOfDesktops < iNoOfDesktops)
		InitNewDesktops(iOldNoOfDesktops, iNoOfDesktops);
}


void CDeskwinEngine::ShowDesktop(TWindowList& aWindowList, TWindow& aForeGroundWindow, TWindowList& aWindowListToShow)
{
	TWindow		window = NULL;
	for (TWindowList::iterator i=aWindowList.begin() ; i!=aWindowList.end(); i++)
	{
		window = *i;
		if (::IsWindow(window))
		{
			if (!::IsWindowVisible(window))
			{
				::ShowWindow(window, SW_SHOW);
				aWindowListToShow.push_back(window);
			}
		}
		else
		{
			WindowDied(window);
		}
	}

	if (::IsWindow(aForeGroundWindow))
	{
		::SetForegroundWindow(aForeGroundWindow);
		::BringWindowToTop(aForeGroundWindow);
	}
}


void CDeskwinEngine::HideDesktop(TWindowList& aWindowList, int aDesktop, TWindow aDragWindow, TWindowList& aWindowListToHide)
{
	for (TWindowList::iterator i=aWindowList.begin() ; i!=aWindowList.end(); i++)
	{
		TWindow		window = *i;
		if (::IsWindow(window))
		{
			if (window == aDragWindow)
			{
				DPRINT((HERE "CDeskwinEngine::HideDesktop"));
			}

			if (!IsStickyWindow(aDesktop, window))
			{
				MyShowWindow(window, false);
				aWindowListToHide.push_back(window);
			}
		}
	}
}


bool StringMatch(const char* aString, const char* aMatchString)
{
	bool match = (recursexp(aString, (char *) aMatchString) >= 0);
	if (!match)
		match = (strstr(aString, aMatchString) != NULL);
	
	return match;
}

static bool IsRuleMatch(const string& aTitleName, const string& aClassName, const TStickyWindow& aStickyWindow) 
{
	bool titleMatch = StringMatch(aTitleName.c_str(), (char *) aStickyWindow.iTitleName.c_str());
	bool classMatch = StringMatch(aClassName.c_str(), (char *) aStickyWindow.iClassName.c_str());

	bool isRuleMatch = (aStickyWindow.iStickyWindowMatch == EStickyWindowMatchAnd) ?
			(titleMatch && classMatch) : (titleMatch || classMatch);

	return isRuleMatch;
}

static bool IsInDesktopList(int aDesktop, const TDesktopList& aDesktopList) 
{
	bool isInDesktopList = (aDesktopList.end() == find(aDesktopList.begin(), aDesktopList.end(), aDesktop)) ? FALSE : TRUE;

	return isInDesktopList;
}


static IsInDesktop(int aDesktop, const TStickyWindow& aStickyWindow) 
{
	bool	isInDesktop = FALSE;
	if (aStickyWindow.iStickyWindowDesktop == EStickyWindowDesktopAll)
	{
		isInDesktop = TRUE;
	}
	else if (aStickyWindow.iStickyWindowDesktop == EStickyWindowDesktopInclude)
	{
		isInDesktop = IsInDesktopList(aDesktop, aStickyWindow.iDesktopList);
	}
	else
	{
		isInDesktop = IsInDesktopList(aDesktop, aStickyWindow.iDesktopList) ? FALSE : TRUE;
	}

	return isInDesktop;
}


void CDeskwinEngine::IsStickyWindow1(int aDesktop, TWindow aWindow, bool& aRuleMatch, bool& aIsSticky, int& aRuleNo)
{
	ASSERT((aDesktop>=-1) && (aDesktop<iNoOfDesktops));

	string	titleName;
	string	className;

	GetWindowName(aWindow, titleName);
	GetClassName(aWindow, className);
	
	int		count	= 0;
	aIsSticky		= aRuleMatch = FALSE;
	aRuleNo			= 0;
	for(TStickyWindowInfo::iterator i = iStickyWindowInfo.begin() ; !aRuleMatch && i!=iStickyWindowInfo.end() ; i++, count++)
	{
		const TStickyWindow&	stickyWindow = *i;
		if (IsRuleMatch(titleName, className, stickyWindow))
		{
			aRuleMatch = TRUE;
			aRuleNo = count;
			if (aDesktop >= 0)
			{
				if (IsInDesktop(aDesktop, stickyWindow))
					aIsSticky = TRUE;
			}
		}
	}
}


bool CDeskwinEngine::IsStickyWindow(int aDesktop, TWindow aWindow)
{
	ASSERT((aDesktop>=0) && (aDesktop<iNoOfDesktops));

	string	titleName;
	string	className;

	GetWindowName(aWindow, titleName);
	GetClassName(aWindow, className);
	
	bool	isSticky = FALSE;
	for(TStickyWindowInfo::iterator i = iStickyWindowInfo.begin() ; !isSticky && i!=iStickyWindowInfo.end() ; i++)
	{
		const TStickyWindow&	stickyWindow = *i;
		if (IsRuleMatch(titleName, className, stickyWindow))
		{
			if (IsInDesktop(aDesktop, stickyWindow))
				isSticky = TRUE;
		}
	}

	return isSticky;
}



void CDeskwinEngine::CaptureDesktop(TWindowList& aWindowList, TWindow& aForeGroundWindow, bool aWorkAroundStupidWindows)
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	aForeGroundWindow = GetForegroundWindow();

		/*
		 * Capture visible windows, ignoring some silly XP helper windows (
		 * when switching desktops we will get rid of them).
		 */
	aWindowList.clear();
	while (window)
	{
		TWindow		nextWindow = ::GetWindow(window, GW_HWNDNEXT);
		if (::IsWindowVisible(window))
		{
			string	classname;
			CDeskwinEngine::GetClassName(window, classname);

			if (IsStupidExplorerWorkerWindow(classname))
			{
				if (aWorkAroundStupidWindows)
					GetRidOfStupidExplorerWorkerWindow(window);
			}
			else if (IsStupidExplorerFaderWindow(classname))
			{
			}
			else if (IsToolTipWindow(classname))
			{
			}
			else
			{
				aWindowList.push_back(window);
			}
		}

		window =  nextWindow;
	}
}

void CDeskwinEngine::MinimiseAll(const TWindow aDeskwinWindow)
{
	TWindow		explorerWindow = GetExplorerWindow();
	TWindow		taskbarWindow = GetTaskbarWindow();
	TWindow		dummy;

	TWindowList currentWindowList;
	CaptureDesktop(currentWindowList, dummy);
	for (TWindowList::iterator i=currentWindowList.begin() ; i != currentWindowList.end() ; i++)
	{
			/*
			 * ignore explorer, taskbar and ourself
			 */
		if (explorerWindow == *i || taskbarWindow == *i || aDeskwinWindow == *i)
			continue;
		::ShowWindow(*i, SW_MINIMIZE);
	}
}

void CDeskwinEngine::WaitUntilWindowsAreUpdated(TWindowList& aWindowListHidden, TWindowList& aWindowListShown)
{
	DPRINT((HERE "WaitUntilWindowsAreUpdated"));

	bool finished = false;
	for (int c=0 ; c<10 && !finished; c++)
	{
			// assume we're all done, then verify all windows that should be hidden are invisible and all
			// all windows that should be shown are visibile
		bool all_ok = true;
		for (TWindowList::iterator i=aWindowListHidden.begin() ; i != aWindowListHidden.end() && all_ok ; i++)
		{
			if (::IsWindowVisible(*i))
			{
				DPRINT((HERE "WaitUntilWindowsAreUpdated: window not yet invisible: 0x%x", *i));
				all_ok = false;
			}
		}
		for (TWindowList::iterator j=aWindowListShown.begin() ; j != aWindowListShown.end() && all_ok ; j++)
		{
			if (!::IsWindowVisible(*j))
			{
				all_ok = false;
				DPRINT((HERE "WaitUntilWindowsAreUpdated: window not yet visible: 0x%x", *j));
			}
		}
		if (!all_ok)
		{
			::Sleep(100);
		}
		else
			finished = true;
	}
	DPRINT((HERE "WaitUntilWindowsAreUpdated"));
}


void CDeskwinEngine::MonitorHungWindows(TWindowList& aWindowListHidden)
{
	for (TWindowList::iterator i=aWindowListHidden.begin() ; i != aWindowListHidden.end() ; i++)
	{
		if (::IsWindowVisible(*i))
		{
				// start thread (close handle strait away - we no longer need it)
			CloseHandle(CreateThread(0, 0, HungThreadWatcher, *i, 0, 0));
		}
	}
}


void CDeskwinEngine::SwitchTo(int aDesktop, TWindow aDragWindow)
{
	if ((aDesktop>=0) && (aDesktop<iNoOfDesktops))
	{
		if (aDesktop != iCurrentDesktop)
		{
			GetRidOfWindowsWhenSwitching();

			TWindowList& currentWindowList = iWindowLists[iCurrentDesktop];
			CaptureDesktop(currentWindowList, iActiveWindows[iCurrentDesktop], true);
			currentWindowList.remove(aDragWindow);

			DPRINT((HERE "********* SwitchTo:: current desktop ************"));
#ifdef _DEBUG
			DumpWindowList(currentWindowList);
#endif

	
			TWindowList		windowListToHide;
			TWindowList		windowListToShow;

			HideDesktop(currentWindowList, aDesktop, aDragWindow, windowListToHide);
			
			iCurrentDesktop = aDesktop;
			ShowDesktop(iWindowLists[iCurrentDesktop], iActiveWindows[iCurrentDesktop], windowListToShow);

			DPRINT((HERE "********* SwitchTo:: new desktop ************"));
#ifdef _DEBUG
			DumpWindowList(iWindowLists[iCurrentDesktop]);
#endif
				// we hide/shown windows asynchronously, so we'll wait a while until they are hidden/shown
			WaitUntilWindowsAreUpdated(windowListToHide, windowListToShow);

				// monitor windows that have hung
			MonitorHungWindows(windowListToHide);

				// finally notify our observers
			for (TDesktopEngineObservers::iterator i=iDesktopEngineObservers.begin() ; i!=iDesktopEngineObservers.end() ; i++)
			{
				MDesktopEngineObserver*	observer = *i;
				observer->HandleDesktopChanged(iCurrentDesktop, aDesktop);
			}
		}
	}
}

int CDeskwinEngine::CurrentDesktop()
{
	return iCurrentDesktop;
}



void CDeskwinEngine::WindowDied(TWindow aWindow)
{
}

void CDeskwinEngine::GetWindowList(int aDesktop, TWindowList& aWindowList)
{
	ASSERT((aDesktop >= 0) && (aDesktop < iNoOfDesktops));

	if (aDesktop == iCurrentDesktop)
	{
		TWindow ignore;
		CaptureDesktop(aWindowList, ignore);
	}
	else
	{
		aWindowList.assign(iWindowLists[aDesktop].begin(), iWindowLists[aDesktop].end());;
	}
}


void CDeskwinEngine::InitNewDesktops(int aOldNoODesktops, int aNewNoODesktops)
{
	TWindow	explorerWindow = GetExplorerWindow();
	TWindow	taskbarWindow = GetTaskbarWindow();

	for (int i=aOldNoODesktops ; i<aNewNoODesktops ; i++)
	{
		TWindowList&	windowList = iWindowLists[i];
		windowList.push_back(taskbarWindow);
		windowList.push_back(explorerWindow);
	}
}


void CDeskwinEngine::GetAllWindows(TWindowList& aWindowList)
{
	TWindow ignore;
	CaptureDesktop(aWindowList, ignore);
	for (int i=0; i<iNoOfDesktops ; i++)
	{
		TWindowList&	windowList1 = iWindowLists[i];
		aWindowList.insert(aWindowList.end(), windowList1.begin(), windowList1.end());
	}

//	DPRINT((HERE "********* GetAllWindows before ************"));
//	DumpWindowList(aWindowList);
	aWindowList.sort();
	aWindowList.unique();
//	DPRINT((HERE "********* GetAllWindows after ************"));
//	DumpWindowList(aWindowList);
}

void CDeskwinEngine::FormatStickyDesktopList(const TStickyWindow& aStickyWindow, string& aString)
{
	bool addList = FALSE;

	switch (aStickyWindow.iStickyWindowDesktop)
	{
		case EStickyWindowDesktopInclude:
			addList = TRUE;
			aString.assign("Only ");
			break;

		case EStickyWindowDesktopExclude:
			addList = TRUE;
			aString.assign("All but ");
			break;
		
		default:
			aString.assign("All");
			break;
	}
	if (addList)
	{
		int			count = 0;
		TCHAR		desktopList[1024];
		TCHAR		num[32];
		desktopList[0] = 0;
		for (TDesktopList::const_iterator i=aStickyWindow.iDesktopList.begin() ; i!=aStickyWindow.iDesktopList.end() ; i++)
//		for (int i=0 ;  i!=aStickyWindow.iDesktopList.size() ; i++)
		{
			int	val = *i;
			sprintf(num, "%i",  val + 1);
			if (count != 0)
				strcat(desktopList, ", ");
			strcat(desktopList, num);
			count++;
		}
		aString.append(desktopList);
	}

}

void CDeskwinEngine::DefaultWindowRule(TWindow aWindow, TStickyWindow& aStickyWindow)
{
	CDeskwinEngine::GetWindowName(aWindow, aStickyWindow.iRuleName);
	CDeskwinEngine::GetWindowName(aWindow, aStickyWindow.iTitleName);
	CDeskwinEngine::GetClassName(aWindow, aStickyWindow.iClassName);
}


void CDeskwinEngine::AddObserver(MDesktopEngineObserver& aDesktopEngineObserver)
{
	iDesktopEngineObservers.push_back(&aDesktopEngineObserver);
}

void CDeskwinEngine::RemoveObserver(MDesktopEngineObserver& aDesktopEngineObserver)
{
	iDesktopEngineObservers.remove(&aDesktopEngineObserver);
}


bool CDeskwinEngine::IsShellWorkerWindowVisible()
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	while (window)
	{
		if (::IsWindowVisible(window))
		{
			string	classname;
			CDeskwinEngine::GetClassName(window, classname);

			if (IsStupidExplorerWorkerWindow(classname))
			{
				return true;
			}
		}
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}
	return false;
}


bool CDeskwinEngine::IsStartMenuVisible()
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	while (window)
	{
		if (::IsWindowVisible(window))
		{
			string	classname;
			CDeskwinEngine::GetClassName(window, classname);

			if (classname.compare(KStartMenuClassName) == 0)
			{
				return true;
			}
		}
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}
	return false;
}

bool CDeskwinEngine::IsShellToolTipVisible()
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
	while (window)
	{
		if (::IsWindowVisible(window))
		{
			string	classname;
			CDeskwinEngine::GetClassName(window, classname);

			if (classname.compare(KToolTipClassName) == 0)
			{
				return true;
			}
		}
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}
	return false;
}

void CDeskwinEngine::GetRidOfWindowsWhenSwitching()
{
	int		count = 5;
	bool	cont;
	POINT	p;
	GetCursorPos(&p);
	do
	{
		count;
		cont = false;
		if (IsShellToolTipVisible())
		{
			DPRINT((HERE "Trying to get rid of tool tip, count=%d", count));
			mouse_event(MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN, -1, -1, 0, NULL);
			mouse_event(MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP, -1, -1, 0, NULL);
			cont = true;
		}
		if (IsStartMenuVisible())
		{
			DPRINT((HERE "Trying to get rid of start menu, count=%d", count));
			mouse_event(MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTDOWN, -1, -1, 0, NULL);
			mouse_event(MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_LEFTUP, -1, -1, 0, NULL);
			cont = true;
		}
		if (cont)
			Sleep(100);
	} while((count-- > 0) && cont);
	SetCursorPos(p.x, p.y);
	DPRINT((HERE "SwitchDesktop, count = %d", count));
}

void CDeskwinEngine::PrepareForExit()
{
	SwitchTo(0);
	SetNoOfDesktops(1);
	DeleteStoredWindowState();
}

/****************************************************
 *
 * Persistency
 *
 ****************************************************/

void CDeskwinEngine::DeleteStoredWindowState()
{
	::RegDeleteKey(HKEY_CURRENT_USER, PERSIST_DESKTOP_BASE);
}

void CDeskwinEngine::RestoreWindowState()
{
	HKEY			ourKey;
	char			name[9];
	unsigned long	szName;
	DWORD			value;
	unsigned long	szValue;
	DWORD			indx;

	if(RegCreateKeyEx(
				HKEY_CURRENT_USER, 
				PERSIST_DESKTOP_BASE, 
				0,
				0,
				0,//REG_OPTION_VOLATILE,
				KEY_ALL_ACCESS,
				NULL,
				&ourKey,
				NULL)	==0) 
	{
		for(indx = 0 ; ; indx++) 
		{
			szName	= 9;
			szValue	= 4;
			if(RegEnumValue(ourKey,indx,name,&szName,0,NULL,(unsigned char *)&value,&szValue)!=0)
				break;

			if(value==0) {
				sscanf(name,"%x",&value);
				MyShowWindow((TWindow)value,true);
//				value=0;
//				RegSetValueEx(ourKey,name,0,REG_DWORD,(unsigned char *)&value,sizeof(value));

			}
		}

		::RegCloseKey(ourKey);
	}
}


void CDeskwinEngine::StoreWindowState(TWindow aWindow, bool aShow)
{
	HKEY	ourKey;
	char	name[9];
	DWORD	value;


	if(RegCreateKeyEx(
			HKEY_CURRENT_USER, 
			PERSIST_DESKTOP_BASE, 
			0, 
			0, 
			0, //REG_OPTION_VOLATILE, 
			KEY_ALL_ACCESS, 
			NULL, 
			&ourKey,
			NULL) ==0) 
	{
		_snprintf(name, 9, "%8.8x", aWindow);
		value = aShow;
		RegSetValueEx(ourKey,name, 0, REG_DWORD, (unsigned char *) &value, sizeof(value));
		RegCloseKey(ourKey);
	}
}



/****************************************
 *
 * DEBUG SECTION
 *
 ****************************************/

#ifdef _DEBUG

void dump_title(const string& s)
{
	DPRINT((HERE "  sticky name=%s", s.c_str()));
}

void dump_class(const string& s)
{
	DPRINT((HERE "  sticky class=%s", s.c_str()));
}

void dump_wnd(const TWindow& window)
{
	string	classname, windowname;
	CDeskwinEngine::GetWindowName(window, windowname);
	CDeskwinEngine::GetClassName(window, classname);
	DPRINT((HERE "  window=0x%08x, name=%s, class=%s", window, windowname.c_str(), classname.c_str()));
}

void dump_stickywhnd(const TWindow& window)
{
	string	classname, windowname;
	CDeskwinEngine::GetWindowName(window, windowname);
	CDeskwinEngine::GetClassName(window, classname);
	DPRINT((HERE "  sticky window=0x%8x, name=%s, class=%s", window, windowname.c_str(), classname.c_str()));
}

void CDeskwinEngine::DumpWindowList(TWindowList& aWindowlist)
{
	for_each(aWindowlist.begin(), aWindowlist.end(), dump_wnd);
}

void CDeskwinEngine::DumpState()
{
	TWindowLists::iterator	i;
	DPRINT((HERE ""));
	DPRINT((HERE "********* DumpState ************"));

	int d = 0;
	for (i = iWindowLists.begin() ; i != iWindowLists.end() ; i++)
	{
		DPRINT((HERE "Desktop %i", d));
				
		DPRINT((HERE "  forground window=0x%8x", iActiveWindows[d]));
		TWindowList&  windowList = *i;
		DumpWindowList(windowList);

		d++;
	}
	DPRINT((HERE "********* DumpState ************"));
	DPRINT((HERE ""));
}




#endif

/*******************************************************************
 *
 * $Log: /Deskwin/Src/CDeskwinEngine.cpp $
 * 
 * 4     10/01/00 9:17 Pese
 * 
 * 3     5/01/00 20:26 Pese
 * 
 * 2     23/12/99 10:16 Pese
 *
 *******************************************************************/
