/*******************************************************************
 *
 *	File:		CDeskwinWindow.h
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 1999, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#ifndef __CDESKWIN_WINDOW_H
#define __CDESKWIN_WINDOW_H

#include "CDeskwinWindowBase.h"
#include "CSmoothWindowMover.h"
#include "CSysTrayIconController.h"
#include "TDeskwinOptions.h"

#ifndef LO_INT16
#define LO_INT16(dw) ((short) (dw & 0xffff))
#endif

#ifndef HI_INT16
#define HI_INT16(dw) ((short) (dw >> 16))
#endif


class CDeskwinEngine;
class CWindowPositionTracker;

#define WM_WINDOW_POS_CHANGED	(WM_USER + 1)
#define WM_WINDOW_CREATED		(WM_USER + 2)
#define WM_SYSTRAY_EVENT		(WM_USER + 3)

class CDesktopNumberWindow : public CWindowEx
{
public:
	CDesktopNumberWindow();
	void Construct();
	void SetDesktop(int aDesktop);
protected:
	long OnWmTimer(WPARAM aWParam, LPARAM aLParam){OnTimer(); return 1;}
	void OnTimer();
	void Paint(CDC& dc);

	CPen	iPen;
	CBrush	iBrush;
};



class CDeskwinWindow : public CDeskwinWindowBase, MSmoothWindowMoverObserver
{
public:
	enum TDockingPlace
	{
		EDockNowhere = 0,
		EDockTop,
		EDockBottom,
		EDockRight,
		EDockLeft,
		EDockLastEntry
	};

	CDeskwinWindow();
	~CDeskwinWindow();
	void Construct(CDeskwinEngine* aDeskwinEngine);

protected:		// state machine
	enum TState
	{
		EStateNormal = 0,
		EStateTooltip,
		EStateDragResizeIntention,		// mouse down in window frame
		EStateDragResize,				// resize the desktop windows
		EStateDragMoreDesktops,			// resize for more/less windows
		EStateDragMoveMyself,			// moving our own window
		EStateDragMoveWindow,			// moving a small window with desktop switching
		EStateFadingOut,				// moving out of the screen
		EStateFadingIn,					// moving back in the screen
		EStateHidden,					// moved outside the screen, only a small rect visible
		EStateLastEntry
	};

	enum TEvent
	{
		EEventLMouseUp = 0,				// from WM_MOUSE and family
		EEventLMouseDown,
		EEventRMouseDown,
		EEventRMouseUp,
		EEventMouseMove,
		EEventSmoothMoveDone,			// smooth window mover has faded us in/out
		EEventTimeout,					// from a WM_TIMER
		EEventPaint,					// from WM_PAINT
		EEventWindowCreated,			// from WM_PAINT
		EEventLastEntry
	};

	TState	iState;						// current state
	void	HandleEvent(TEvent aEvent);
	
		/*
		 * for each state and event we have an event handler. The default
		 * will be ignore event.
		 */
	typedef void (CDeskwinWindow::*TEventHandler)(TEvent aEvent);
	TEventHandler	iEventHandlers[EStateLastEntry][EEventLastEntry];

	void DefaultEventHandlers();
	void SetEventHandler(TState aState, TEvent aEvent, TEventHandler aEventHandler);
	void SetEventHandlerForHiddenState();

		/*
		 * All event handlers
		 */
	void IgnoreEvent(TEvent aEvent);
	void NormalPaint(TEvent aEvent);
	void ResizeNoOfDesktopsPaint(TEvent aEvent);
	void ResizePaint(TEvent aEvent);
	void StartDragging(TEvent aEvent);
	void DoDragResizeIntention(TEvent aEvent);
	void DragResizeDone(TEvent aEvent);
	void DoDragResize(TEvent aEvent);
	void DragMoveMyselfDone(TEvent aEvent);
	void DoDragMoveMyself(TEvent aEvent);
	void DragMoveWindowDone(TEvent aEvent);
	void DoDragMoveWindow(TEvent aEvent);
	void StartDragMoreDesktops();
	void DragMoreDesktopsDone(TEvent aEvent);
	void DoDragMoreDesktops(TEvent aEvent);
	void NormalMouseDown(TEvent aEvent);
	void NormalMouseUp(TEvent aEvent);
	void NormalMouseMouse(TEvent aEvent);
	void StartFadingOut(TEvent aEvent);
	void FadingOutDone(TEvent aEvent);
	void FadingInDone(TEvent aEvent);
	void StartFadingIn(TEvent aEvent);
	void ShowPopupMenu(TEvent aEvent);
	void WindowCreated(TEvent aEvent);

	union 
	{
		struct
		{
			int		key;
			int		x;
			int		y;
		} iMouse;

		CDC*	iPaintCDC;

		TWindow		iWindowCreated;
	} iEvent;


protected:		// window messages
		// raw window messages
	long OnWmLButtonDown(WPARAM aWParam, LPARAM aLParam){OnLButtonDown(aWParam, LO_INT16(aLParam), HI_INT16(aLParam)); return 1;}
	long OnWmLButtonUp(WPARAM aWParam, LPARAM aLParam){OnLButtonUp(aWParam, LO_INT16(aLParam), HI_INT16(aLParam)); return 1;}
	long OnWmRButtonDown(WPARAM aWParam, LPARAM aLParam){OnRButtonDown(aWParam, LO_INT16(aLParam), HI_INT16(aLParam)); return 1;}
	long OnWmRButtonUp(WPARAM aWParam, LPARAM aLParam){OnRButtonUp(aWParam, LO_INT16(aLParam), HI_INT16(aLParam)); return 1;}
	long OnWmMouseMove(WPARAM aWParam, LPARAM aLParam){OnMouseMove(aWParam, LO_INT16(aLParam), HI_INT16(aLParam)); return 1;}
	long OnWmCaptureChanged(WPARAM aWParam, LPARAM aLParam){OnCaptureChanged(); return 1;}
	long OnWmDisplayChanged(WPARAM aWParam, LPARAM aLParam){OnDisplayChanged(aWParam, LOWORD(aLParam), HIWORD(aLParam)); return 1;}
	long OnWmTimer(WPARAM aWParam, LPARAM aLParam){OnTimer(); return 1;}
	long OnWmHotkey(WPARAM aWParam, LPARAM aLParam){OnHotkey(aWParam, LOWORD(aLParam), HIWORD(aLParam)); return 1;}
	long OnWmWindowPosChanged(WPARAM aWParam, LPARAM aLParam){OnWindowPosChanged(); return 1;}
	long OnWmWindowCreated(WPARAM aWParam, LPARAM aLParam){OnWindowCreated((TWindow) aWParam); return 1;}
	long OnWmSysTrayEvent(WPARAM wParam, LPARAM lParam){OnSysTrayEvent((int) lParam); return 1;}



		// translated window messages
		// Without exception they all save state in iEvent and then call HandleEvent
//	void Paint(CDC dc);
	virtual void DoPaint(CDC& dc);
	void OnRButtonDown(int fwKeys, int aX, int aY);
	void OnRButtonUp(int fwKeys, int aX, int aY);
	void OnLButtonDown(int fwKeys, int aX, int aY);
	void OnLButtonUp(int fwKeys, int aX, int aY);
	void OnMouseMove(int fwKeys, int aX, int aY);
	void OnCaptureChanged();
	void OnDisplayChanged(int aBitsPerPixel, int aX, int aY);
	void OnTimer();
	void OnHotkey(int aHotkeyId, int aModifier, int aVirtKey);
	void OnWindowPosChanged();
	void OnWindowCreated(TWindow aWindow);
	void OnSysTrayEvent(int aParam);

	virtual bool	IsStickyWindow(int aDesktop, TWindow aWindow);

protected:		// from CWindow
	virtual void OnQueryEndSession();

protected:			// popup menu and its command handlers
	CMenu		iMenu;

	void DoCmdQuit();
	void DoCmdAppearanceOptions();
	void DoCmdDockOptions();
	void DoCmdAutoHideOptions();
	void DoCmdHotkeyOptions();
	void DoCmdColorOptions();
	void DoCmdStickyWindows();
	void DoCmdAbout();
	void DoCmdReadme();

protected:		// implementation
	enum TResizeArea			// the four corners of the window
	{
		EResizeAreaNone = 0,
		EResizeAreaTopLeft,
		EResizeAreaTopRight,
		EResizeAreaBottomLeft,
		EResizeAreaBottomRight,
	};

	virtual void Destroy();
	void ShowOptionsMenu(int aX, int aY);

//	void ShrinkScreenRectToDesktopRect(RECT& rect, bool aClip = true);
//	void InflateDesktopRectToScreenRect(RECT& rect);
	bool PointInDesktop(int aClientX, int aClientY, int& aDesktop, int& aDesktopX, int& aDesktopY);
	bool PointInWindowFrame(int aClientX, int aClientY);
	TResizeArea GetResizeArea(int aX, int aY);

	void UpdateWindowForDesktopSwitch(int aNewDesktop, int& aX, int& aY);
	void ExcludeFrameRect(RECT& aRect);
	void InvalidateDragWindowBackground(RECT aScreenRect);
	void RepaintDesktop(int aDesktop, bool aRedrawBackground = false);

	void PaintDesktop(CDC& dc, int aDesktop, bool aIsActive);
//	void PaintFrames(CDC& aDc, RECT& aRect);
	void DoResizeNoOfDesktopsPaint(CDC& aDc, RECT& aRect);
	
	void DockWindow(TDockingPlace aDockingPlace);
	void StartDraggingMyself(int aX, int aY);
	void StopDraggingMyself();
	void StartDraggingWindow(int aX, int aY);
	void StopDraggingWindow(bool aCancelled = false);
	void StartDragResize();

	TDockingPlace GetDockingPlace();

	void GetDesktopRect(int aDesktop, RECT& aRect);

	TWindow FindWindowInDesktop(int aDesktop, int aX, int aY);
	
	void StartAutoHideTimer();
	void ResetAutoHideTimer();

	void SwitchDesktop(int aNewDesktop, TWindow aDragWindow = NULL);

	void	SetOptions();

	void CheckStickyWindows();
	void CheckStickyWindow(TWindow aWindow);

	void SetZOrder();
protected:

//	CDesktopNumberWindow*	iDesktopNumberWindow;

	HICON	iWindowIcon;
	int		iActiveDesktop;

	TResizeArea	iDragResizeArea;
	int		iDragStartSizeX;
	int		iDragStartSizeY;
	int		iDragStartX;
	int		iDragStartY;
	RECT	iDragRect;
	bool	iDragOldAlwaysOnTop;

	int		iDragHorDesktops;
	int		iDragVertDesktops;

	TWindow	iDragWindow;

	BOOL	iDragWindowMaximized;
	RECT	iDragWindowOrigialRect;

	POINT	iFadeMove;

	bool	iIsDocked;
	TDockingPlace		iPreviousDockingPlace;

	CDeskwinEngine*			iDeskwinEngine;
	CSmoothWindowMover*		iSmoothWindowMover;
	CWindowPositionTracker* iWindowPositionTracker;
	CSysTrayIconController*	iSysTrayIconController;

	HANDLE					iCursorNeSw;
	HANDLE					iCursorNwSe;
	HANDLE					iCursorMove;
	HANDLE					iCursorMoreDesktops;
	HANDLE					iPreviousCursor;
	HANDLE					iCurrentCursor;

	bool					iMenuAlreadyRunning;
protected:		// implements MSmoothWindowMoverObserver
	virtual void DoneSmoothWindowMoving(CWindow* aWindow);

protected:		// state that will be saved/restored
	int				iStatePosX;
	int				iStatePosY;
	TDockingPlace	iStateDockingPlace;

		/*
		 * To/from registry
		 */
	void RestoreState();
	void SaveState();
	void SaveOptions();
	
	void SetColorOptions();
	void GetColorOptions();
	void GetDeskwinBaseOptions();
	void SetDeskwinBaseOptions();

protected:		// options
	TDeskwinOptions		iOptions;

	void RegisterHotkeys(TDeskwinOptions& aOptions);
	void UnregisterHotkeys(TDeskwinOptions& aOptions);
};


/*******************************************************************
 *
 * $Log: /Deskwin/Src/CDeskwinWindow.h $
 * 
 * 4     10/01/00 9:17 Pese
 * 
 * 3     5/01/00 20:26 Pese
 * 
 * 2     23/12/99 10:16 Pese
 *
 *******************************************************************/


#endif			/* __CDESKWIN_WINDOW_H */