/*******************************************************************
 *
 *	File:		CWindowPositionTracker.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 1999, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include "Deskwin.h"
#include <process.h>
#include "CWindowPositionTracker.h"


/*
 * keep STL happy
 */
bool operator <(const TWindowPosInfo& aWindowPosInfo1, const TWindowPosInfo& aWindowPosInfo2)
{
	return ((int) &aWindowPosInfo1 < (int) &aWindowPosInfo2);
}

bool operator ==(const TWindowPosInfo& aWindowPosInfo1, const TWindowPosInfo& aWindowPosInfo2)
{
	return ((int) &aWindowPosInfo1 == (int) &aWindowPosInfo2);
}


/*
 * Position tracker determines whether or not the currently visible set of
 * toplevel windows have changed position (moved, minimised, closed, new window
 * etc). It runs in a background thread and is effectively taking snapshots every
 * x ms (x specified by the client). The snapshot consists of the set of windows
 * and their current position. If the current snapshot is different from the
 * previous one, it sends a message (message and target window specified by 
 * client) to report the change.
 * To prevent active hogging of the CPU, the position tracker can run at two
 * polling frequencies: f1 and f2, both specified by the client. Initially, the 
 * position tracker polls at frequency f1. If at some point in time it detects a 
 * window position change, it will start polling at frequency f2. It will do so as 
 * long as two subsequent snaptshot are the same; after that it polls at f1 again.
 * The client can control the position trackers thread, by start-stop and pause-resume.
 *
 * The thread control is trivial. We use two events: one for pause-resume and the
 * one to stop. The latter means the thread should terminate. The thread handle is
 * used to wait until the thread has gone. The two events and the thread handle
 * are used in the context of the client's thread to control the worker thread.
 */

CWindowPositionTracker::CWindowPositionTracker()
	: iReportWindow(0), iReportMessage(0), iActiveTrackingTime(10000), iBackgroundTrackingTime(10000), iIsFastTracking(false), iPrivateThread(0)
{
	iPauseEvent		= CreateEvent(NULL, true, false, NULL);		// manual reset
	iStopEvent		= CreateEvent(NULL, true, false, NULL);		// manual reset
}


CWindowPositionTracker::~CWindowPositionTracker()
{
	StopWindowPositionTracking();

	CloseHandle(iPauseEvent);
	CloseHandle(iStopEvent);
}


void CWindowPositionTracker::StartWindowPositionTracking(TWindow aWindowToReportTo, int aMessageToReport, int aActiveTrackingTime, int aBackgroundTrackingTime)
{
	DPRINT((HERE "CWindowPositionTracker::StartWindowPositionTracking"));

	iReportWindow = aWindowToReportTo;
	iReportMessage = aMessageToReport; 
	iActiveTrackingTime = aActiveTrackingTime;
	iBackgroundTrackingTime = aBackgroundTrackingTime;

	if (iPrivateThread == 0)
	{
		iPrivateThread = (HANDLE) _beginthread(&CWindowPositionTracker::KickoffWindowTracker, 0, this);
		if (iPrivateThread == (HANDLE) -1)
		{
			DPRINT((HERE "CWindowPositionTracker::StartWindowPositionTracking, thread NOT started"));
			iPrivateThread = 0;
		}
	}
}


void CWindowPositionTracker::KickoffWindowTracker(void* aContext)
{
	CWindowPositionTracker*		windowPositionTracker = (CWindowPositionTracker*) aContext;

	windowPositionTracker->DoTheWindowTracking();

	DPRINT((HERE "CWindowPositionTracker::KickoffWindowTracker done"));
	_endthread();		// bye bye
}


void CWindowPositionTracker::StopWindowPositionTracking()
{
	if (IsThreadRunning())
	{

			/*
			 * Wakeup thread and instruct it to stop and...
			 */
		ResumeWindowPositionTracking();
		SetEvent(iStopEvent);
		
			/*
			 * ... wait a couple seconds (5) for that to happen
			 */
		if (::WaitForSingleObject(iPrivateThread, 5000) == WAIT_OBJECT_0)
		{
			DPRINT((HERE "window tracking thread successfully stopped"));
		}
		else
		{
			DPRINT((HERE "Hmm, window tracking wasn't stopped in 5 seconds"));
		}
	}
	iPrivateThread = 0;
}


void CWindowPositionTracker::PauseWindowPositionTracking()
{
	DPRINT((HERE "CWindowPositionTracker::PauseWindowPositionTracking"));

	if (IsThreadRunning())
		ResetEvent(iPauseEvent);
}


void CWindowPositionTracker::ResumeWindowPositionTracking()
{
//	DPRINT((HERE "CWindowPositionTracker::ResumeWindowPositionTracking"));
	if (IsThreadRunning())
		SetEvent(iPauseEvent);
}


void CWindowPositionTracker::TakeSnapShot(TWindowPosInfos& aWindowPosInfo)
{
	TWindow window		= ::GetWindow(::GetDesktopWindow(), GW_CHILD);
	
		/*
		 * Capture info about visible windows
		 */
	aWindowPosInfo.clear();
	while (window)
	{
		if (::IsWindowVisible(window))
		{
			TWindowPosInfo	windowPosInfo;

			windowPosInfo.iWindow = window;
			::GetWindowRect(window, &windowPosInfo.iPosition);
			aWindowPosInfo.push_back(windowPosInfo);
		}
		window =  ::GetWindow(window, GW_HWNDNEXT);
	}
}

bool CWindowPositionTracker::AreSnapshotsDifferent(TWindowPosInfos& aWindowPosInfo1, TWindowPosInfos& aWindowPosInfo2)
{
	if (aWindowPosInfo1.size() != aWindowPosInfo2.size())
		return true;

	for (TWindowPosInfos::iterator i=aWindowPosInfo1.begin(), j=aWindowPosInfo2.begin(); i!=aWindowPosInfo1.end() && j!=aWindowPosInfo2.end() ; i++, j++)
	{
		TWindowPosInfo&		windowPosInfo1 = *i;
		TWindowPosInfo&		windowPosInfo2 = *j;

		if (windowPosInfo1.iWindow != windowPosInfo2.iWindow)
			return true;

		if ((windowPosInfo1.iWindow == windowPosInfo2.iWindow) && !EqualRect(&windowPosInfo1.iPosition, &windowPosInfo2.iPosition))
			return true;
	}
	return false;
}


void CWindowPositionTracker::DoTheWindowTracking()
{
	TWindowPosInfos		positions[2];
	int					oldpos, newpos;

	oldpos = 0;
	newpos = 1;

		/*
		 * Take snapshots until requested to die
		 */
	TakeSnapShot(positions[oldpos]);
	do
	{
			/*
			 * If someone paused us, we will sleep forever until we're continued
			 */
		::WaitForSingleObject(iPauseEvent, INFINITE);

			/*
			 * Sleep inbetween snapshots at frequency f1 or f2
			 */
		::Sleep(iIsFastTracking ? iActiveTrackingTime : iBackgroundTrackingTime);

		TakeSnapShot(positions[newpos]);

//		DPRINT((HERE "Comparing snapshots...."));

		if (AreSnapshotsDifferent(positions[oldpos], positions[newpos]))
		{
			iIsFastTracking = true;
			::PostMessage(iReportWindow, iReportMessage, 0, 0);
		}
		else
		{
			iIsFastTracking = false;
		}

		swap(oldpos, newpos);
		
	} while (::WaitForSingleObject(iStopEvent, 0) != WAIT_OBJECT_0);
}

/*******************************************************************
 *
 * $Log: /Deskwin/Src/CWindowPositionTracker.cpp $
 * 
 * 4     10/01/00 9:17 Pese
 * 
 * 3     5/01/00 20:26 Pese
 * 
 * 2     23/12/99 10:16 Pese
 *
 *******************************************************************/
