/*  Copyright (C) 1993   Marc Stern  (internet: stern@mble.philips.be)  */

#include "strings.h"
#include <stdlib.h>
#define near 

/*
   Functions :   matchset
                 match
                 recursexp
                 regexp
*/



/***
 *  Function    :  matchset
 *
 *  Description :  Test if a character matches a set expression.
 *
 *  Parameters  :  in    char c          character to be matched
 *                 in    char *pattern   regular expression to match
 *
 *  Parameters  :  in   char   *string
 *
 *  Decisions   :  The following symbols are treated specially:
 *
 *              \  quote next character      -  range of values
 *              ^  non-inclusion (if first character)
 *
 *              ex: aeiou0-9   match a, e, i, o, u, and 0 thru 9
 *                  ^aeiou0-9  match anything but a, e, i, o, u, and 0 thru 9
 *
 *  Return      :  1 or 0
 *
 *  OS/Compiler :  All
 ***/

int matchset( char c, char *pattern )

{
 const char *begin;
 int match_ok = 1;

 if ( ! c ) return 0;

 if ( *pattern == '^' )
    {
     match_ok = 0;
     pattern ++;
    }

 for ( begin = pattern; *pattern; begin = pattern++ )
    {
     if ( (*pattern == '-') && (pattern != begin) )  /* range */
        {
         pattern ++; if ( *pattern == '\\' ) pattern ++;
         if ( (min(*pattern, *begin) <= c) && (max(*pattern, *begin) >= c) )
            return match_ok;
        }

     if ( *pattern == '\\' ) pattern++;

     if ( *pattern == c ) return match_ok;
    }

 return (! match_ok);
}




/***
 *  Function    :  match (internal)
 *
 *  Description :  Returns the number of character of a string matched
 *                 by a one character regular expression.
 *
 *  Decisions   :  The following symbols are treated specially:
 *
 *                 .  any character             \  quote next character
 *                 [] set of characters
 *
 *  Parameters  :  in   char *string    input string to be matched
 *                 in   char *pattern   regular expression to match
 *
 *  Side-effects:  pattern contents will be destroyed.
 *
 *  Return      :  number of character matched by regular expression
 *                 0 if not matched
 *
 *  OS/Compiler :  All
 ***/

static int near match( const char *string, char *pattern )
{
 char *ptr;
 int  length = 0;

 switch ( *pattern )
 {
  case '.' : *pattern = '\0';
             length = strlen(string);
             break;

  case '[' : for ( ptr = ++pattern; *ptr ; ptr ++ )
                if ( *ptr == ']' )
                   {
                    int i;
                    for ( i = 0; ; i++ )
                       if ( *(ptr - i - 1) != '\\' ) break;

                    if ( ! (i % 2) ) break; /* not "\]" */
                   }

                if ( *ptr ) *ptr = '\0';
                while ( matchset(*string++, pattern) ) length ++;
                break;

  case '\\': pattern ++;

  default  : while ( *string++ == *pattern ) length ++;
                      *pattern = '\0';
                      break;
 }

 return length;
}




/***
 *  Function    :  recursexp
 *
 *  Description :  Returns the number of character of a string matched
 *                 by a regular expression.
 *
 *  Decisions   :  The following symbols are treated specially:
 *
 *                 .  any character             \  quote next character
 *                 *  match zero or more        +  match one or more
 *                 ?  match zero or one         [] set of characters
 *
 *
 *  Parameters  :  in   char *string    input string to be matched
 *                 in   char *pattern   regular expression to match
 *
 *  Return      :  number of character matched by regular expression
 *                 -1 if not matched
 *
 *  OS/Compiler :  All
 ***/

int recursexp( const char *string, char *pattern )
{
 int count1, count2 = -1, minone = 0;
 char *pattrn, *ptr;

 if ( ! *pattern ) return 0;

 ptr = pattrn = strdup( pattern );     /* store it for free() */
 count1 = match( string, pattrn );
 if ( count1 < 0 )
    {
     free( ptr );
     return -1;
    }

 while ( *pattrn++ );

 switch ( *pattrn )
 {
  case '\0': free( ptr );
             if ( count1 ) return 1;
                      else return -1;

  case '*': pattrn++;
            minone = 0;
            break;

  case '+': if ( ! count1 )
               {
                free( ptr );
                return -1;
               }
             pattrn++;
             break;

  case '?': count1 = min( count1, 1 );
            pattrn++;
            break;

  default : if ( ! count1 )
               {
                free( ptr );
                return -1;
               }
            count1 = 1;
            break;
 }

 for ( ; count1 >= 0; count1 -- )
    {
     count2 = recursexp( string + count1, pattrn );
     if ( count2 >= 0 ) break;
    }

 free( ptr );

 if ( count2 < 0 ) return -1;
 if ( ! count1 && minone ) return -1;

 return (count1 + count2);
}




/***
 *  Function    :  regexp
 *
 *  Description :  Returns the string matched by a regular expression
 *                 into a string.
 *
 *  Decisions   :  The following symbols are treated specially:
 *
 *                 ^  start of line             $  end of line
 *                 .  any character             \  quote next character
 *                 *  match zero or more        +  match one or more
 *                 ?  match zero or one         [] set of characters
 *        
 *              ex: [aeiou0-9]   match a, e, i, o, u, and 0 thru 9
 *                  [^aeiou0-9]  match anything but a, e, i, o, u, and 0 thru 9
 *
 *  Parameters  :  out   char *outstr    resulting string
 *                 in    char *string    input string in which we search
 *                 in    char *pattern   regular expression to match
 *
 *  Return      :  - NULL if not found
 *                 - pointer to resulting string
 *                 - if ( outstr == NULL ) returns pointer to matched string
 *                   inside 'string'.
 *
 *  OS/Compiler :  All
 ***/

char *regexp( char *outstr, const char *string, const char *pattern )
{
 char *ptr, *pattrn;                                                
 int count, begin = 0, end = 0, length;

 pattrn = strdup( pattern );
 count = strlen( pattrn ) - 1;
 ptr = pattrn + count;
 if ( (*ptr == '$') )   /* Match end of line ? */
    {
     int i;
     for ( i = 0; i < count ; i++ )
        if ( *(ptr - i - 1) != '\\' ) break;

     if ( ! (i % 2) )   /* Match end of line */
        {
         end = 1;
         *ptr = '\0';
        }
    }

 ptr = pattrn;  /* store it for free() */

 if ( *pattrn == '^' )        /* Match begin of line */
    {
     begin = 1;
     pattrn ++;
    }

 for ( length = strlen(string); *string; string++, length-- )
    {
     count = recursexp( string, pattrn );
     if ( begin ) break;  /* only valid as first characters */
     if ( count < 0 ) continue;   /* not found */
     if ( ! end || count == length ) break;
    }

 free( ptr );

 if ( count < 0 ||
      (end && count != length)
    )
    {
     if ( outstr ) *outstr = '\0';
     return NULL;
    }

 if ( outstr )
    {
     strleft( outstr, string, count );
     return outstr;
    }

 return (char *)string;
}


#ifdef TEST

#include <stdio.h>
#include <stdlib.h>

void main( void )
{
 char string[255], pattern[255], result[255];

 for (;;) {
           printf( "\n    String : " );
           if ( ! *gets(string) ) exit(0);
           printf( "    Pattern: " ); gets( pattern );

           regexp( result, string, pattern );
           printf( "\n    Result : %s\n\n", result );
          }
}

#endif

#include "left.c"


/* 
 * Part of ACTLib developed by Marc Stern.
 *
  ACTLIB:  Another C Tools LIBrary (plus TurboVision tools)
  Copyright (C) 1993   Marc Stern  (internet: stern@mble.philips.be)

  This package contains several libraries and their source code.
  Some functions are quite complex, some other are very trivial but
  may be useful just to centralize frequently called functions
  (and so allows to coherently propagate modifications).
  The goal of this package is to provide useful tools for developers
  and also to serve as example codes either for beginners programmers
  or for non-novice ones that want to use some low-level tricks
  of MS-DOS like:
     - Intra-application Communications
     - get extended keyboard keys (F1,... Alt-X,...)
     - put a key back into keyboard buffer
     - open more than 20 files
     - get the density of a floppy drive
     - test the availability of a drive (exist, formatted,...)
     - very fast file copy
     - very fast multiple files deletion
     - input a password
     - get/set disk volume label
     - get cpu type (86, 186, 286, 386, 486)
     - disallow debugging of your program

  Most of the code should be compiler- and operating system-independant
  but has only been tested under Borland C 3.1 and Microsoft C 7.0
  compilers (MS-DOS).
  Some functions have been intensively tested and are used in commercial
  products, some have been developed a long time ago and never been
  tested since. There may left some bugs (even important ones)

  All these sources have been either developed or collected (and modified)
  by Marc Stern.

  I will not be responsible for any damage due to the use of these libraries.
  This code is provided as-is WITHOUT ANY WARRANTY.

  You can freely distribute them, modify them and use them for any purpose
  (even include it into commercial products).

  If you find any bugs or interesting modifications please report them to
  me; they will so be included in the next public releases.

  If you use these libraries, please tell it to me.
  I will so try to maintain and enhance them.

  ####################################################################
  #                                                                  # 
  #   Also available:            ***   TVTOOLS   ***                 #
  #                                                                  #
  #     Free library for Turbo-Vision containing a lot of addings.   #
  #     Available as TVTOOLxx.ZIP (same version number as ACTLIB).   #
  #                                                                  #
  ####################################################################
                  

  Files description:
  -----------------

  *.LIB:  libraries compiled with Borland C++ 3.1 (overlay compatible)

  DATE.LIB    : general-purpose date & time functions
  STRINGS.LIB : general-purpose string handling functions
  TOOLS.LIB   : miscellanous functions (most of all for MS-DOS only)
  BCTOOLS.LIB : user input/output related functions (dedicated to Turbo C)

  DBGALLOC.C  : replacement for malloc/realloc/free functions
                to perform heap memory bounds-checking
                (for MS-DOS only).
   

  *.H   : include files
  *.ZIP : source files


  See include files for descriptions of the functions.

  How to use/compile the package:  
  ------------------------------
  My functions generally call some other functions from the
  'general' libraries (STRINGS.LIB, TOOLS.LIB,...).
  This means that you have to include the related include files
  and to link (some of) the other libraries with your application.


  Special thanks:
  ==============

  1.   To my company: S.A. Philips Industrial and Telecommunication Systems
       (some of the code was developed for business project).

  2.   To all people that written code that I modified or that I used
       to understand some processes.
       As I don't have their names anymore I thanks them all in one shot.


  Who am I?
  =========
          
  Just a few words about me...

  I finished my studies in 1989 (Civil Engineer in computering) and
  I am working since for a division of Philips - one of the greatest
  specialists in software security (local and transfer confidentiality,
  authentication, coherency, DES/RSA/DSS cyphering,...) and voice recognition.

  I worked about six months in Software Engineering (under UNIX and VMS),
  one year PC support (LAN organisation,...), one year in Quality Assurance
  (Beta-test,...) and I am now working in the security area.

  I also write the great XSET program that allows you to easily write
  efficient batch files. See file XSET.TXT to have more details.

----------------------------------------------------------------------------
                    
Stern Marc          S.A. Philips Industrial and Telecommunication Systems
                    Rue des Deux Gares 80
                    1070 Bruxelles
                    Belgium

                    Fax:  +32-2 525.66.00
                    Tel:  +32-2 525.63.32 (9h - 16h Medium European Time)

E-mail: stern@mble.philips.be (internet)
        >INTERNET:stern@mble.philips.be (compuserve)
*/