/*******************************************************************
 *
 *	Module:		regutil.c
 *
 *	Creator:	Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 1999, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <windows.h>
#include <tchar.h>
#include <string.h>

#include "regutil.h"


/*************************************
 *
 * STATICS
 *
 *************************************/

static long SetRegKeyAnyValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	IN	DWORD				ValueType,
	IN	const void			*ValueData,
	IN	int					ValueDataSize
	)
{
	HKEY		hKey;
	long		Status;
	
	Status = RegOpenKeyEx(
					RootKey, 
					KeyName,
					0,
					KEY_WRITE, 
					&hKey
					);
	if (Status == ERROR_SUCCESS)
	{
		Status = RegSetValueEx( 
						hKey,
						ValueName,
						0, 
						ValueType, 
						ValueData,
						ValueDataSize
						);
		RegCloseKey(hKey);			// ignore return code
	}
	else
	{
		return Status;
	}

	return Status;
}


static long GetRegKeyAnyValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	OUT	DWORD				*ValueType,
	OUT	void				*ValueData,
	IN OUT	int				*ValueDataSize
	)
{
	HKEY		hKey;
	long		Status;
	
	Status = RegOpenKeyEx(
					RootKey, 
					KeyName,
					0,
					KEY_READ, 
					&hKey
					);
	if (Status == ERROR_SUCCESS)
	{
		Status = RegQueryValueEx( 
						hKey,
						ValueName,
						0, 
						ValueType, 
						(BYTE *) ValueData,
						ValueDataSize
						);
		RegCloseKey(hKey);			// ignore return code
	}
	else
	{
		return Status;
	}

	return Status;
}


/*************************************
 *
 * INTERFACE ROUTINES
 *
 *************************************/


long GetRegKeyUintValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	OUT	unsigned int		*UintValue
	)
{
	DWORD	error, Type, IntegerSize=sizeof(unsigned int);
	int		LocalIntegerValue;

	error = GetRegKeyAnyValue(
								RootKey,
								KeyName,
								ValueName,
								&Type,
								&LocalIntegerValue,
								&IntegerSize
								);
	if (!error)
	{
		if (Type == REG_DWORD)
		{
			*UintValue = LocalIntegerValue;
		}
		else
		{
			error = ERROR_INVALID_DATA;
		}
	}
	return error;
}


long GetRegKeyIntegerValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	OUT	int					*IntegerValue
	)
{
	DWORD	error, Type, IntegerSize=sizeof(int);
	int		LocalIntegerValue;

	error = GetRegKeyAnyValue(
								RootKey,
								KeyName,
								ValueName,
								&Type,
								&LocalIntegerValue,
								&IntegerSize
								);
	if (!error)
	{
		if (Type == REG_DWORD)
		{
			*IntegerValue = LocalIntegerValue;
		}
		else
		{
			error = ERROR_INVALID_DATA;
		}
	}
	return error;
}



long GetRegKeyStringValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	OUT	_TCHAR				*StringValue,
	IN OUT int				*StringValueLen
	)
{
	DWORD	error, Type;

	error = GetRegKeyAnyValue(
								RootKey,
								KeyName,
								ValueName,
								&Type,
								StringValue,
								StringValueLen
								);
	if (!error)
	{
		if (!Type == REG_SZ)
		{
			error = ERROR_INVALID_DATA;
		}
	}
	return error;
}


long DestroyRegKey(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*SubKeyName
	)
{
	HKEY		hKey;
	long		Status;
	
	Status = RegOpenKeyEx(
					RootKey, 
					KeyName,
					0,
					KEY_WRITE, 
					&hKey
					);
	if (Status == ERROR_SUCCESS)
	{
		char	subSubKeyName[1024];
		HKEY	hSubKey;

		Status = RegOpenKeyEx(
						hKey, 
						SubKeyName,
						0,
						KEY_WRITE | KEY_READ, 
						&hSubKey
						);

		if (Status == ERROR_SUCCESS)
		{
			while (RegEnumKey(
					hSubKey, 
					0,
					subSubKeyName, 
					1024 
					) == ERROR_SUCCESS)
			{
				if (DestroyRegKey(hSubKey, _T(""), subSubKeyName) != ERROR_SUCCESS)
					break;
				
			}
			RegCloseKey(hSubKey);
		}

		Status = RegDeleteKey(hKey, SubKeyName);
		
		RegCloseKey(hKey);			// ignore return code
	}
	else
	{
		return Status;
	}

	return Status;
}


long EnumRegKey(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	int					Index,
	OUT	_TCHAR				*SubKeyName,
	OUT	int					SubKeyNameLen
	)
{
	HKEY		hKey;
	long		Status;
	
	Status = RegOpenKeyEx(
					RootKey, 
					KeyName,
					0,
					KEY_READ, 
					&hKey
					);
	if (Status == ERROR_SUCCESS)
	{
		Status = RegEnumKey(
						hKey, 
						Index, 
						SubKeyName, 
						SubKeyNameLen
						); 
		RegCloseKey(hKey);			// ignore return code
	}
	else
	{
		return Status;
	}

	return Status;
}


long CreateRegKey(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*SubKeyName
	)
{
	HKEY		hKey, hSubKey;
	long		Status;
	DWORD		Disposition;
	
	Status = RegOpenKeyEx(
					RootKey, 
					KeyName,
					0,
					KEY_WRITE, 
					&hKey
					);
	if (Status == ERROR_SUCCESS)
	{
		Status = RegCreateKeyEx(
						hKey, 
						SubKeyName, 
						0, 
						_TEXT(""), 
						REG_OPTION_NON_VOLATILE, 
						KEY_WRITE, 
						NULL, 
						&hSubKey, 
						&Disposition
						); 		
		RegCloseKey(hSubKey);		// ignore return code
		RegCloseKey(hKey);			// ignore return code
	}
	else
	{
		return Status;
	}

	return Status;
}


long SetRegKeyStringValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	IN	const _TCHAR		*StringValue
	)
{
	return SetRegKeyAnyValue(
				RootKey,
				KeyName,
				ValueName,
				REG_SZ,
				StringValue,
				_tcslen(StringValue) + 1
				);
}

long SetRegKeyIntegerValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	IN	int					IntegerValue
	)
{
	return SetRegKeyAnyValue(
				RootKey,
				KeyName,
				ValueName,
				REG_DWORD,
				&IntegerValue,
				sizeof(int)
				);
}


long SetRegKeyUintValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName,
	IN	unsigned int		UintValue
	)
{
	return SetRegKeyAnyValue(
				RootKey,
				KeyName,
				ValueName,
				REG_DWORD,
				&UintValue,
				sizeof(unsigned int)
				);
}




long DestroyRegKeyStringValue(
	IN	HKEY				RootKey,
	IN	const _TCHAR		*KeyName,
	IN	const _TCHAR		*ValueName
	)
{
	HKEY		hKey;
	long		Status;
	
	Status = RegOpenKeyEx(
					RootKey, 
					KeyName,
					0,
					KEY_WRITE, 
					&hKey
					);
	if (Status == ERROR_SUCCESS)
	{
		Status = RegDeleteValue( 
						hKey,
						ValueName
						);
		RegCloseKey(hKey);			// ignore return code
	}
	else
	{
		return Status;
	}

	return Status;
}



/*******************************************************************
 *
 * $Log: /Deskwin/Src/Regutil.c $
 * 
 * 4     10/01/00 9:17 Pese
 * 
 * 3     5/01/00 20:26 Pese
 * 
 * 2     23/12/99 10:16 Pese
 *
 *******************************************************************/
